from ir_measures import measures
from .base import Measure, ParamInfo, SumAgg


class _NumQ(measures.Measure):
    """
    The total number of queries.
    """
    __name__ = 'NumQ'
    NAME = __name__
    SUPPORTED_PARAMS = {}

    def aggregator(self):
        return SumAgg()


NumQ = _NumQ()
measures.register(NumQ)



from ir_measures import measures
from .base import Measure, ParamInfo


class _Bpref(measures.Measure):
    """
    Binary Preference (Bpref).
    This measure examines the relative ranks of judged relevant and non-relevant documents. Non-judged documents are not considered. 

::

    @inproceedings{Buckley2004RetrievalEW,
      title={Retrieval evaluation with incomplete information},
      author={Chris Buckley and Ellen M. Voorhees},
      booktitle={SIGIR},
      year={2004}
    }
    """
    __name__ = 'Bpref'
    NAME = __name__
    SUPPORTED_PARAMS = {
        'rel': measures.ParamInfo(dtype=int, default=1, desc='minimum relevance score to be considered relevant (inclusive)')
    }


Bpref = _Bpref()
BPref = Bpref
measures.register(Bpref, ['BPref'])



registry = {}
def register(provider):
	registry[provider.NAME] = provider
	return provider

from .base import Provider, Evaluator
from .fallback_provider import FallbackProvider
from .compat_provider import CompatProvider
from .cwl_eval import CwlEvalProvider, CwlMetric
from .pyndeval_provider import PyNdEvalProvider
from .pytrec_eval_provider import PytrecEvalProvider
from .judged_provider import JudgedProvider
from .gdeval_provider import GdevalProvider
from .trectools_provider import TrectoolsProvider
from .msmarco_provider import MsMarcoProvider



import os
import unittest
import itertools
import ir_measures
from ir_measures import *


class TestCompat(unittest.TestCase):

    def test_measures(self):
        qrels = list(ir_measures.read_trec_qrels(os.path.join(os.path.dirname(__file__), 'compat.qrels')))
        run = list(ir_measures.read_trec_run(os.path.join(os.path.dirname(__file__), 'compat.run')))
        provider = ir_measures.compat
        # based on a manual execution of https://github.com/claclark/Compatibility
        expected_results = [
            [Compat(p=0.95), [('31_1', 0.51779512165509), ('31_2', 0.018400100569017922)]],
            [Compat(p=0.9), [('31_1', 0.3761334522946854), ('31_2', 0.004344079941789211)]],
            [Compat(p=0.8), [('31_1', 0.16723008845234535), ('31_2', 0.00022806427320561776)]],
        ]
        for measure, expected in expected_results:
            with self.subTest(measure=measure):
                self.assertTrue(provider.supports(measure))
                results = list(provider.iter_calc([measure], qrels, run))
                for result, (query_id, value) in zip(results, expected):
                    self.assertAlmostEqual(result.query_id, query_id, delta=1e-9)
                    self.assertAlmostEqual(result.value, value, delta=1e-9)
        self.assertAlmostEqual(provider.calc_aggregate([Compat(p=0.95)], qrels, run)[Compat(p=0.95)], 0.268097611, delta=1e-9)


if __name__ == '__main__':
    unittest.main()



from ir_measures import measures
from .base import Measure, ParamInfo


class _ERR(measures.Measure):
    """
    The Expected Reciprocal Rank (ERR) is a precision-focused measure.
    In essence, an extension of reciprocal rank that encapsulates both graded relevance and
    a more realistic cascade-based user model of how users brwose a ranking.
    """
    __name__ = 'ERR'
    NAME = __name__
    SUPPORTED_PARAMS = {
        'cutoff': measures.ParamInfo(dtype=int, required=False, desc='ranking cutoff threshold'),
    }


ERR = _ERR()
measures.register(ERR)



from ir_measures import measures
from .base import Measure, ParamInfo


class _P(measures.Measure):
    """
    Basic measure for that computes the percentage of documents in the top cutoff results
    that are labeled as relevant. cutoff is a required parameter, and can be provided as
    P@cutoff.

::

    @misc{rijsbergen:1979:ir,
      title={Information Retrieval.},
      author={Van Rijsbergen, Cornelis J},
      year={1979},
      publisher={USA: Butterworth-Heinemann}
    }
    """
    __name__ = 'P'
    NAME = __name__
    SUPPORTED_PARAMS = {
        'cutoff': measures.ParamInfo(dtype=int, required=True, desc='ranking cutoff threshold'),
        'rel': measures.ParamInfo(dtype=int, default=1, desc='minimum relevance score to be considered relevant (inclusive)')
    }


P = _P()
Precision = P
measures.register(P, ['Precision'])



import sys
import contextlib
import ir_measures
from ir_measures import providers, measures
from ir_measures.providers.base import Any, Choices, Metric, NOT_PROVIDED


class PyNdEvalProvider(providers.Provider):
    """
    pyndeval
    """
    NAME = 'pyndeval'
    SUPPORTED_MEASURES = [
        measures._ERR_IA(cutoff=Any(), rel=Any(), judged_only=Any()),
        measures._nERR_IA(cutoff=Any(), rel=Any(), judged_only=Any()),
        measures._alpha_DCG(cutoff=Any(), alpha=Any(), rel=Any(), judged_only=Any()),
        measures._alpha_nDCG(cutoff=Any(), alpha=Any(), rel=Any(), judged_only=Any()),
        measures._NRBP(alpha=Any(), beta=Any(), rel=Any()),
        measures._nNRBP(alpha=Any(), beta=Any(), rel=Any()),
        measures._AP_IA(rel=Any(), judged_only=Any()),
        measures._P_IA(cutoff=Any(), rel=Any(), judged_only=Any()),
        measures._StRecall(cutoff=Any(), rel=Any()),
    ]

    def __init__(self):
        super().__init__()
        self.pyndeval = None

    def _evaluator(self, measures, qrels):
        measures = ir_measures.util.flatten_measures(measures)

        qrels = [self._map_qrel_namedtuple(q) for q in ir_measures.util.QrelsConverter(qrels).as_namedtuple_iter()]

        # Depending on the measures, we may need multiple invocations of pyndeval
        # (e.g., with different rel_level, alpha, and beta)
        invokers = self._build_invokers(measures, qrels)

        if all(not inv.evaluator.has_multiple_subtopics('any') for inv in invokers):
            sys.stderr.write('WARNING: All queries have only 1 subtopic! The results from this metric are probably not '
                             'valid. Make sure that you are using diversity qrels, and that they are provided as a '
                             'dataframe (with iteration column) or an iterable of GenericQrel (with iteration).\n')

        return PyNdEvalEvaluator(measures, qrels, invokers)

    def _build_invokers(self, measures, qrels):
        DEFAULT_ALPHA, DEFAULT_BETA = 0.5, 0.5
        invocations = {}
        for measure in measures:
            if measure.NAME in ('NRBP', 'nNRBP'):
                invocation_key = (measure['rel'], measure['alpha'], measure['beta'], False)
                measure_str = f'{measure.NAME}'
            elif measure.NAME in ('alpha_DCG', 'alpha_nDCG'):
                invocation_key = (measure['rel'], measure['alpha'], DEFAULT_BETA, measure['judged_only'])
                measure_str = f'{measure.NAME.replace("_", "-")}@{measure["cutoff"]}'
            elif measure.NAME in ('ERR_IA', 'nERR_IA', 'P_IA'):
                invocation_key = (measure['rel'], DEFAULT_ALPHA, DEFAULT_BETA, measure['judged_only'])
                measure_str = f'{measure.NAME.replace("_", "-")}@{measure["cutoff"]}'
            elif measure.NAME == 'AP_IA':
                invocation_key = (measure['rel'], DEFAULT_ALPHA, DEFAULT_BETA, measure['judged_only'])
                measure_str = f'MAP-IA'
            elif measure.NAME == 'StRecall':
                invocation_key = (measure['rel'], DEFAULT_ALPHA, DEFAULT_BETA, False)
                measure_str = f'strec@{measure["cutoff"]}'
            else:
                raise ValueError(f'unsupported measure {measure}')

            if invocation_key not in invocations:
                invocations[invocation_key] = {}
            invocations[invocation_key][measure_str] = measure

        invokers = []
        for (rel_level, alpha, beta, judged_only), measure_map in invocations.items():
            invokers.append(PyNdEvalInvoker(self.pyndeval, qrels, measure_map, rel_level, alpha, beta, judged_only))

        return invokers

    def initialize(self):
        try:
            import pyndeval
            self.pyndeval = pyndeval
        except ImportError as ex:
            raise RuntimeError('pyndeval not available', ex)

    def _map_qrel_namedtuple(self, record):
        # Map iteration to subtopic_id. Since it's technically optional, fall back on
        # the default value if not provided. There is a warning later in the process
        # if there are not multiple subtopics found.
        subtopic_id = record.iteration if 'iteration' in record._fields else '0'
        return self.pyndeval.SubtopicQrel(
            query_id=record.query_id,
            subtopic_id=subtopic_id,
            doc_id=record.doc_id,
            relevance=record.relevance)


class PyNdEvalEvaluator(providers.Evaluator):
    def __init__(self, measures, qrels, invokers):
        query_ids = {q.query_id for q in qrels}
        super().__init__(measures, query_ids)
        self.invokers = invokers

    def _iter_calc(self, run):
        # Convert run to dict_of_dict
        run = ir_measures.util.RunConverter(run).as_sorted_namedtuple_iter()
        for invoker in self.invokers:
            yield from invoker.iter_calc(run)


class PyNdEvalInvoker:
    def __init__(self, pnd, qrels, measure_map, rel_level, alpha, beta, judged_only):
        self.evaluator = None
        self.evaluator = pnd.RelevanceEvaluator(qrels, measure_map.keys(), relevance_level=rel_level, alpha=alpha, beta=beta)
        self.measure_map = measure_map
        self.qid_did_filter = None
        if judged_only:
            self.qid_did_filter = set((qrel.query_id, qrel.doc_id) for qrel in qrels)

    def iter_calc(self, run):
        if self.qid_did_filter is not None: # used when judged_only
            filtered_run = {}
            for qid in run:
                filtered_run[qid] = {}
                for did, score in run[qid].items():
                    if (qid, did) in self.qid_did_filter:
                        filtered_run[qid][did] = score
            run = filtered_run
        for record in self.evaluator.evaluate_iter(run):
            query_id = record['query_id']
            del record['query_id']
            for measure_str, value in record.items():
                yield Metric(query_id=query_id, measure=self.measure_map[measure_str], value=value)


providers.register(PyNdEvalProvider())



import io
import pkgutil
import subprocess
import tempfile
import contextlib
import ir_measures
from ir_measures import providers, measures, Metric
from ir_measures.providers.base import Any, Choices, NOT_PROVIDED


class GdevalProvider(providers.Provider):
    """
    gdeval
    """
    NAME = 'gdeval'
    SUPPORTED_MEASURES = [
        measures._nDCG(cutoff=Any(required=True), dcg=Choices('exp-log2')),
        measures._ERR(cutoff=Any(required=True)),
    ]

    def _evaluator(self, measures, qrels):
        MEASURES = ('nDCG', 'ERR')
        cutoffs = {}
        for measure in ir_measures.util.flatten_measures(measures):
            if measure.NAME in MEASURES:
                cutoff = measure['cutoff']
                if cutoff not in cutoffs:
                    cutoffs[cutoff] = [None] * len(MEASURES)
                cutoffs[cutoff][MEASURES.index(measure.NAME)] = measure
            else:
                raise ValueError(f'unsupported measure {measure}')
        invocations = []
        for cutoff, (NDCG, ERR) in cutoffs.items():
            invocations.append((cutoff, NDCG, ERR))
        qrels = list(ir_measures.util.QrelsConverter(qrels).as_namedtuple_iter())
        return GdevalEvaluator(measures, qrels, invocations)

    def initialize(self):
        try:
            subprocess.check_output(['perl', '--version'])
        except CalledProcessError as ex:
            raise RuntimeError('perl not available', ex)


class GdevalEvaluator(providers.Evaluator):
    def __init__(self, measures, qrels, invocations):
        super().__init__(measures, set(q.query_id for q in qrels))
        self.qrels = qrels
        self.invocations = invocations

    def _iter_calc(self, run):
        with tempfile.NamedTemporaryFile() as perlf, \
             ir_measures.util.QrelsConverter(self.qrels).as_tmp_file() as qrelsf, \
             ir_measures.util.RunConverter(run).as_tmp_file() as runf:
            perlf_contents = pkgutil.get_data('ir_measures', 'bin/gdeval.pl')
            perlf.write(perlf_contents)
            perlf.flush()
            for cutoff, nDCG_measure, ERR_measure in self.invocations:
                cmd = ['perl', perlf.name, qrelsf.name, runf.name, str(cutoff)]
                output = subprocess.check_output(cmd)
                output = output.decode().replace('\t', ' ').split('\n')
                for i, s in enumerate(output):
                    if s == '' or i == 0:
                        continue
                    arr = s.split(',')
                    assert len(arr) == 4
                    _, qid, ndcg, err = arr
                    if nDCG_measure is not None:
                        yield Metric(query_id=qid, measure=nDCG_measure, value=float(ndcg))
                    if ERR_measure is not None:
                        yield Metric(query_id=qid, measure=ERR_measure, value=float(err))


providers.register(GdevalProvider())



from ir_measures import measures
from .base import Measure, ParamInfo


class _Rprec(measures.Measure):
    """
    The precision of at R, where R is the number of relevant documents for a given query. Has the cute property that
    it is also the recall at R.

::

    @misc{Buckley2005RetrievalSE,
      title={Retrieval System Evaluation},
      author={Chris Buckley and Ellen M. Voorhees},
      annote={Chapter 3 in TREC: Experiment and Evaluation in Information Retrieval},
      howpublished={MIT Press},
      year={2005}
    }
    """
    __name__ = 'Rprec'
    NAME = __name__
    SUPPORTED_PARAMS = {
        'rel': measures.ParamInfo(dtype=int, default=1, desc='minimum relevance score to be considered relevant (inclusive)')
    }


Rprec = _Rprec()
RPrec = Rprec
measures.register(Rprec, ['RPrec'])



import warnings
import re
import io
import ast
import contextlib
import itertools
import tempfile
from typing import List
from collections import namedtuple
from typing import NamedTuple, Union
import ir_measures

class Qrel(NamedTuple):
    query_id: str
    doc_id: str
    relevance: int
    iteration: str = '0'

class ScoredDoc(NamedTuple):
    query_id: str
    doc_id: str
    score: float

class Metric(NamedTuple):
    query_id: str
    measure: 'Measure'
    value: Union[float, int]


class GenericQrel(Qrel):
    def __new__(self, *args, **kwargs):
        warnings.warn("GenericQrel deprecated in 0.2.0. Please use ir_measures.Qrel instead.", DeprecationWarning)
        return super().__new__(self, *args, **kwargs)
GenericQrel._fields = Qrel._fields


class GenericScoredDoc(ScoredDoc):
    def __new__(self, *args, **kwargs):
        warnings.warn("GenericScoredDoc deprecated in 0.2.0. Please use ir_measures.ScoredDoc instead.", DeprecationWarning)
        return super().__new__(self, *args, **kwargs)
GenericScoredDoc._fields = ScoredDoc._fields


class QrelsConverter:
    def __init__(self, qrels):
        self.qrels = qrels
        self._predicted_format = None

    def tee(self, count):
        t, err = self.predict_type()
        if t == 'namedtuple_iter':
            teed_qrels = itertools.tee(self.qrels, count)
            return [QrelsConverter(qrels) for qrels in teed_qrels]
        return [self for _ in range(count)]

    def predict_type(self):
        if self._predicted_format:
            return self._predicted_format
        result = 'UNKNOWN'
        error = None
        if isinstance(self.qrels, dict):
            result = 'dict_of_dict'
        elif hasattr(self.qrels, 'itertuples'):
            cols = self.qrels.columns
            missing_cols = [f for f in Qrel._fields if f not in cols and f not in Qrel._field_defaults]
            if not missing_cols:
                result = 'pd_dataframe'
            else:
                error = f'DataFrame missing columns: {list(missing_cols)} (found {list(cols)})'
        elif hasattr(self.qrels, '__iter__'):
            # peek
            # TODO: is this an OK approach?
            self.qrels, peek_qrels = itertools.tee(self.qrels, 2)
            sentinal = object()
            item = next(peek_qrels, sentinal)
            if isinstance(item, tuple) and hasattr(item, '_fields'):
                fields = item._fields
                missing_fields = [f for f in Qrel._fields if f not in fields and f not in Qrel._field_defaults]
                if not missing_fields:
                    result = 'namedtuple_iter'
                else:
                    error = f'namedtuple iter missing fields: {list(missing_fields)} (found {list(fields)})'
            elif item is sentinal:
                result = 'namedtuple_iter'
            else:
                error = f'iterable not a namedtuple iterator'
        else:
            required_fields = tuple(f for f in Qrel._fields if f not in Qrel._field_defaults)
            error = f'unexpected format; please provide either: (1) an iterable of namedtuples (fields {required_fields}, e.g., from ir_measures.Qrel); (2) a pandas DataFrame with columns {required_fields}; or (3) a dict-of-dict'
        self._predicted_format = (result, error)
        return result, error

    def as_dict_of_dict(self):
        t, err = self.predict_type()
        if t == 'dict_of_dict':
            return self.qrels
        else:
            result = {}
            for qrel in self.as_namedtuple_iter():
                if qrel.query_id not in result:
                    result[qrel.query_id] = {}
                result[qrel.query_id][qrel.doc_id] = qrel.relevance
            return result

    def as_namedtuple_iter(self):
        t, err = self.predict_type()
        if t == 'namedtuple_iter':
            yield from self.qrels
        if t == 'dict_of_dict':
            for query_id, docs in self.qrels.items():
                for doc_id, relevance in docs.items():
                    yield Qrel(query_id=query_id, doc_id=doc_id, relevance=relevance)
        if t == 'pd_dataframe':
            if 'iteration' in self.qrels.columns:
                yield from (Qrel(qrel.query_id, qrel.doc_id, qrel.relevance, qrel.iteration) for qrel in self.qrels.itertuples())
            else:
                yield from (Qrel(qrel.query_id, qrel.doc_id, qrel.relevance) for qrel in self.qrels.itertuples())
        if t == 'UNKNOWN':
            raise ValueError(f'unknown qrels format: {err}')

    def as_pd_dataframe(self):
        t, err = self.predict_type()
        if t == 'pd_dataframe':
            if 'iteration' not in self.qrels.columns:
                return self.qrels.assign(iteration=['0'] * len(self.qrels))
            return self.qrels
        else:
            pd = ir_measures.lazylibs.pandas()
            return pd.DataFrame(self.as_namedtuple_iter())

    @contextlib.contextmanager
    def as_tmp_file(self):
        with tempfile.NamedTemporaryFile(mode='w+t') as f:
            for qrel in self.as_namedtuple_iter():
                f.write('{query_id} 0 {doc_id} {relevance}\n'.format(**qrel._asdict()))
            f.flush()
            f.seek(0)
            yield f




class RunConverter:
    def __init__(self, run):
        self.run = run
        self._predicted_format = None

    def tee(self, count):
        t, err = self.predict_type()
        if t == 'namedtuple_iter':
            teed_run = itertools.tee(self.run, count)
            return [RunConverter(run) for run in teed_run]
        return [self for _ in range(count)]

    def predict_type(self):
        if self._predicted_format:
            return self._predicted_format
        result = 'UNKNOWN'
        error = None
        if isinstance(self.run, dict):
            result = 'dict_of_dict'
        elif hasattr(self.run, 'itertuples'):
            cols = self.run.columns
            missing_cols = set(ScoredDoc._fields) - set(cols)
            if not missing_cols:
                result = 'pd_dataframe'
            else:
                error = f'DataFrame missing columns: {list(missing_cols)} (found {list(cols)})'
        elif hasattr(self.run, '__iter__'):
            # peek
            # TODO: is this an OK approach?
            self.run, peek_run = itertools.tee(self.run, 2)
            sentinal = object()
            item = next(peek_run, sentinal)
            if isinstance(item, tuple) and hasattr(item, '_fields'):
                fields = item._fields
                missing_fields = set(ScoredDoc._fields) - set(fields)
                if not missing_fields:
                    result = 'namedtuple_iter'
                else:
                    error = f'namedtuple iter missing fields: {list(missing_fields)} (found {list(fields)})'
            elif item is sentinal:
                result = 'namedtuple_iter'
            else:
                error = f'iterable not a namedtuple iterator'
        else:
            error = f'unexpected format; please provide either: (1) an iterable of namedtuples (fields {ScoredDoc._fields}, e.g., from ir_measures.ScoredDoc); (2) a pandas DataFrame with columns {ScoredDoc._fields}; or (3) a dict-of-dict'
        self._predicted_format = (result, error)
        return result, error

    def as_dict_of_dict(self):
        t, err = self.predict_type()
        if t == 'dict_of_dict':
            return self.run
        else:
            result = {}
            for scored_doc in self.as_namedtuple_iter():
                if scored_doc.query_id not in result:
                    result[scored_doc.query_id] = {}
                result[scored_doc.query_id][scored_doc.doc_id] = scored_doc.score
            return result

    def as_namedtuple_iter(self):
        t, err = self.predict_type()
        if t == 'namedtuple_iter':
            yield from self.run
        if t == 'dict_of_dict':
            for query_id, docs in self.run.items():
                for doc_id, score in docs.items():
                    yield ScoredDoc(query_id=query_id, doc_id=doc_id, score=score)
        if t == 'pd_dataframe':
            yield from (ScoredDoc(sd.query_id, sd.doc_id, sd.score) for sd in self.run.itertuples())
        if t == 'UNKNOWN':
            raise ValueError(f'unknown run format: {err}')

    def as_sorted_namedtuple_iter(self):
        qid = None
        items = []
        def flush():
            return sorted(items, key=lambda x: x.score, reverse=True)
        for item in self.as_namedtuple_iter():
            if qid is None or item.query_id != qid:
                yield from flush()
                items = []
                qid = item.query_id
            items.append(item)
        if qid is not None:
            yield from flush()

    def as_pd_dataframe(self):
        t, err = self.predict_type()
        if t == 'pd_dataframe':
            return self.run
        else:
            pd = ir_measures.lazylibs.pandas()
            return pd.DataFrame(self.as_namedtuple_iter())

    @contextlib.contextmanager
    def as_tmp_file(self):
        with tempfile.NamedTemporaryFile(mode='w+t') as f:
            ranks = {}
            for scoreddoc in self.as_namedtuple_iter():
                key = scoreddoc.query_id
                rank = ranks.setdefault(key, 0)
                f.write('{query_id} Q0 {doc_id} {rank} {score} run\n'.format(**scoreddoc._asdict(), rank=rank))
                ranks[key] += 1
            f.flush()
            f.seek(0)
            yield f


def parse_trec_qrels(file):
    warnings.warn("parse_trec_qrels deprecated in 0.2.0. Please use ir_measures.read_trec_qrels() instead.", DeprecationWarning)
    return read_trec_qrels(file)


def read_trec_qrels(file):
    if hasattr(file, 'read'):
        for line in file:
            if line.strip():
                query_id, iteration, doc_id, relevance = line.split()
                yield Qrel(query_id=query_id, doc_id=doc_id, relevance=int(relevance), iteration=iteration)
    elif isinstance(file, str):
        if '\n' in file:
            yield from read_trec_qrels(io.StringIO(file))
        else:
            with open(file, 'rt') as f:
                yield from read_trec_qrels(f)

def parse_trec_run(file):
    warnings.warn("parse_trec_run deprecated in 0.2.0. Please use ir_measures.read_trec_run() instead.", DeprecationWarning)
    return read_trec_run(file)

def read_trec_run(file):
    if hasattr(file, 'read'):
        for line in file:
            if line.strip():
                query_id, iteration, doc_id, rank, score, tag = line.split()
                yield ScoredDoc(query_id=query_id, doc_id=doc_id, score=float(score))
    elif isinstance(file, str):
        if '\n' in file:
            yield from read_trec_run(io.StringIO(file))
        else:
            with open(file, 'rt') as f:
                yield from read_trec_run(f)

def flatten_measures(measures):
    result = set()
    for measure in measures:
        if isinstance(measure, ir_measures.measures.MultiMeasures):
            result = result | measure.measures
        else:
            result.add(measure)
    return result


_AST_PARSE_ERROR = 'problem parsing measure {}; must be in format Measure(k1=v1, k2=v2)@c'


def _ast_to_value(node):
    if isinstance(node, ast.Num):
        return node.n
    if isinstance(node, ast.Str):
        return node.s
    if isinstance(node, ast.NameConstant):
        return node.value
    raise ValueError(_AST_PARSE_ERROR.format('values must be str, float, int, bool, etc.'))


def parse_measure(measure: str) -> 'Measure':
    if isinstance(measure, ir_measures.Measure):
        return measure
    try:
        node = ast.parse(measure).body
    except SyntaxError as e:
        raise ValueError(_AST_PARSE_ERROR.format(str(e)))
    if len(node) != 1:
        raise ValueError(_AST_PARSE_ERROR.format('multiple expressions'))
    node = node[0]
    if not isinstance(node, ast.Expr):
        raise ValueError(_AST_PARSE_ERROR.format('not an expression'))
    node = node.value
    at_param = None
    args = {}
    if isinstance(node, ast.BinOp) and isinstance(node.op, ast.MatMult):
        at_param = _ast_to_value(node.right)
        node = node.left
    if isinstance(node, ast.Call):
        if len(node.args):
            raise ValueError(_AST_PARSE_ERROR.format('args must be named'))
        for keyword in node.keywords:
            args[keyword.arg] = _ast_to_value(keyword.value)
        node = node.func
    if not isinstance(node, ast.Name):
        raise ValueError(_AST_PARSE_ERROR.format('unexpected expression'))
    measure_name = node.id
    if measure_name not in ir_measures.measures.registry:
        raise NameError(f'measure not found: {measure_name}')
    measure = ir_measures.measures.registry[measure_name]
    if at_param is not None:
        args[measure.AT_PARAM] = at_param
    return measure(**args)


def parse_trec_measure(measure: str) -> List['Measure']:
    TREC_NAME_MAP = {
        'ndcg': (ir_measures.nDCG, None, None),
        'P': (ir_measures.P, 'cutoff', [5, 10, 15, 20, 30, 100, 200, 500, 1000]),
        'map_cut': (ir_measures.AP, 'cutoff', [5, 10, 15, 20, 30, 100, 200, 500, 1000]),
        'Rndcg': (None, None, None),
        'num_nonrel_judged_ret': (None, None, None),
        'set_map': (ir_measures.SetAP, None, None),
        '11pt_avg': (None, None, None),
        'relative_P': (None, None, None),
        'gm_map': (None, None, None),
        'binG': (None, None, None),
        'set_P': (ir_measures.SetP, None, None),
        'infAP': (ir_measures.infAP, None, None),
        'bpref': (ir_measures.Bpref, None, None),
        'num_rel_ret': (ir_measures.NumRelRet, None, None),
        'ndcg_rel': (None, None, None),
        'recip_rank': (ir_measures.RR, None, None),
        'recall': (ir_measures.R, 'cutoff', [5, 10, 15, 20, 30, 100, 200, 500, 1000]),
        'set_recall': (ir_measures.SetR, None, None),
        'utility': (None, None, None),
        'set_relative_P': (ir_measures.SetRelP, None, None),
        'num_ret': (ir_measures.NumRet, None, None),
        'num_rel': (ir_measures.NumRel, None, None),
        'ndcg_cut': (ir_measures.nDCG, 'cutoff', [5, 10, 15, 20, 30, 100, 200, 500, 1000]),
        # 'runid': (None, None, None), # not really a measure
        'Rprec': (ir_measures.Rprec, None, None),
        'Rprec_mult': (None, None, None),
        'relstring': (None, None, None),
        'gm_bpref': (None, None, None),
        'num_q': (ir_measures.NumQ, None, None),
        'map': (ir_measures.AP, None, None),
        'G': (None, None, None),
        'success': (ir_measures.Success, 'cutoff', [1, 5, 10]),
        'set_F': (ir_measures.SetF, 'beta', [1.]),
        'iprec_at_recall': (ir_measures.IPrec, 'recall', [0.0, 0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9, 1.0]),
    }
    import pytrec_eval
    if measure in pytrec_eval.supported_nicknames:
        result = []
        skipped = []
        for sub_name in sorted(pytrec_eval.supported_nicknames[measure]):
            try:
                result += parse_trec_measure(sub_name)
            except ValueError:
                if sub_name != 'runid':
                    skipped.append(sub_name)
        if skipped:
            print(f'skipped {skipped}: measures not yet supported')
        return result

    # Adapted from <https://github.com/cvangysel/pytrec_eval/blob/master/py/__init__.py#L76>
    RE_BASE = r'{}[\._]([0-9]+(\.[0-9]+)?(,[0-9]+(\.[0-9]+)?)*)'

    # break apart measures in any of the following formats and combine
    #  1) meas          -> {meas: {}}  # either non-parameterized measure or use default params
    #  2) meas.p1       -> {meas: {p1}}
    #  3) meas_p1       -> {meas: {p1}}
    #  4) meas.p1,p2,p3 -> {meas: {p1, p2, p3}}
    #  5) meas_p1,p2,p3 -> {meas: {p1, p2, p3}}
    if measure in TREC_NAME_MAP:
        meas, arg_name, default_args = TREC_NAME_MAP[measure]
        if meas is None:
            raise ValueError(f'known trec name {measure}, but not yet supported')
        if arg_name is not None and default_args is not None:
            result = []
            for arg in default_args:
                result.append(meas(**{arg_name: arg}))
            return result
        return [meas]
    matches = ((m, re.match(RE_BASE.format(re.escape(m)), measure)) for m in TREC_NAME_MAP)
    match = next(filter(lambda x: x[1] is not None, matches), None)
    if match is None:
        raise ValueError('unkonwn measure {}'.format(measure))
    base_meas, meas_args = match[0], match[1].group(1)
    meas, arg_name, _ = TREC_NAME_MAP[base_meas]
    if meas is None:
        raise ValueError(f'known trec name {measure}, but not yet supported')
    if arg_name is None:
        raise ValueError(f'unknown argument for {measure}: {meas_args}')
    result = []
    dtype = meas.SUPPORTED_PARAMS[arg_name].dtype
    for arg in meas_args.split(','):
        result.append(meas(**{arg_name: dtype(arg)}))
    return result


def convert_trec_name(measure: str) -> List['Measure']:
    warnings.warn("convert_trec_name deprecated in 0.2.0. Please use ir_measures.parse_trec_measure() instead.", DeprecationWarning)
    return parse_trec_measure(measure)



from ir_measures import measures
from .base import Measure, ParamInfo


class _RBP(measures.Measure):
    """
    The Rank-Biased Precision (RBP).

::

     @article{Moffat:2008:RPM:1416950.1416952,
       author = {Moffat, Alistair and Zobel, Justin},
       title = {Rank-biased Precision for Measurement of Retrieval Effectiveness},
       journal = {ACM Trans. Inf. Syst.},
       year = {2008},
       url = {http://doi.acm.org/10.1145/1416950.1416952}
     }
    """
    __name__ = 'RBP'
    NAME = __name__
    SUPPORTED_PARAMS = {
        'cutoff': measures.ParamInfo(dtype=int, required=False, desc='ranking cutoff threshold'),
        'p': measures.ParamInfo(dtype=float, default=0.8, desc='persistence'),
        'rel': measures.ParamInfo(dtype=int, required=False, desc='minimum relevance score to be considered relevant (inclusive), or NOT_PROVIDED to use graded relevance')
    }


RBP = _RBP()
measures.register(RBP)



from ir_measures import measures
from .base import Measure, ParamInfo


class _nDCG(measures.Measure):
    """
    The normalized Discounted Cumulative Gain (nDCG).
    Uses graded labels - systems that put the highest graded documents at the top of the ranking.
    It is normalized wrt. the Ideal NDCG, i.e. documents ranked in descending order of graded label.

::

    @article{Jarvelin:2002:CGE:582415.582418,
      author = {J\"{a}rvelin, Kalervo and Kek\"{a}l\"{a}inen, Jaana},
      title = {Cumulated Gain-based Evaluation of IR Techniques},
      journal = {ACM Trans. Inf. Syst.},
      volume = {20},
      number = {4},
      year = {2002},
      pages = {422--446},
      numpages = {25},
      url = {http://doi.acm.org/10.1145/582415.582418},
    }
    """
    __name__ = 'nDCG'
    NAME = __name__
    SUPPORTED_PARAMS = {
        'cutoff': measures.ParamInfo(dtype=int, required=False, desc='ranking cutoff threshold'),
        'dcg': measures.ParamInfo(dtype=str, choices=['log2', 'exp-log2'], default='log2', desc='DCG formulation')
    }


nDCG = _nDCG()
NDCG = nDCG
measures.register(nDCG, ['NDCG'])



from ir_measures import measures
from .base import Measure, ParamInfo


class _Judged(measures.Measure):
    """
    Percentage of results in the top k (cutoff) results that have relevance judgments. Equivalent to P@k with
    a rel lower than any judgment.
    """
    __name__ = 'Judged'
    NAME = __name__
    SUPPORTED_PARAMS = {
        'cutoff': measures.ParamInfo(dtype=int, required=True, desc='ranking cutoff threshold'),
    }


Judged = _Judged()
measures.register(Judged)



import logging
import sys
from typing import NamedTuple, Union
from cwl.cwl_eval import TrecQrelHandler, RankingMaker
from cwl.ruler.cwl_ruler import CWLRuler, PrecisionCWLMetric, RRCWLMetric, APCWLMetric, RBPCWLMetric, BPMCWLMetric, NDCGCWLMetric, NERReq8CWLMetric, NERReq9CWLMetric, NERReq10CWLMetric, NERReq11CWLMetric, INSTCWLMetric, INSQCWLMetric
import ir_measures
from ir_measures import providers, measures, Metric
from ir_measures.providers.base import Any, Choices, NOT_PROVIDED
import logging

logger = logging.getLogger('ir_measures.cwl_eval')
logger.setLevel('WARNING')


class CwlMetric(NamedTuple):
    query_id: str
    measure: 'Measure'
    value: Union[float, int] # i.e., expected_utility
    expected_total_utility: float
    expected_cost: float
    expected_total_cost: float
    expected_items: float


class CwlEvalProvider(providers.Provider):
    """
    cwl_eval, providing C/W/L ("cool") framework measures.

    https://github.com/ireval/cwl

::

    @inproceedings{azzopardi2019cwl,
      author = {Azzopardi, Leif and Thomas, Paul and Moffat, Alistair},
      title = {cwl\\_eval: An Evaluation Tool for Information Retrieval},
      booktitle = {SIGIR},
      year = {2019}
    }
    """
    NAME = 'cwl_eval'
    SUPPORTED_MEASURES = [
        measures._P(cutoff=Any(), rel=Any()),
        measures._RR(cutoff=Choices(NOT_PROVIDED), rel=Any()),
        measures._AP(cutoff=Choices(NOT_PROVIDED), rel=Any()),
        measures._RBP(cutoff=Choices(NOT_PROVIDED), rel=Any(required=True), p=Any()),
        measures._BPM(cutoff=Any(), T=Any(), min_rel=Any(), max_rel=Any(required=True)),
        measures._SDCG(cutoff=Any(required=True), dcg=Choices('log2'), min_rel=Any(), max_rel=Any(required=True)),
        measures._NERR8(cutoff=Any(required=True), min_rel=Any(), max_rel=Any(required=True)),
        measures._NERR9(cutoff=Any(required=True), min_rel=Any(), max_rel=Any(required=True)),
        measures._NERR10(p=Any(), min_rel=Any(), max_rel=Any(required=True)),
        measures._NERR11(T=Any(), min_rel=Any(), max_rel=Any(required=True)),
        measures._INST(T=Any(), min_rel=Any(), max_rel=Any(required=True)),
        measures._INSQ(T=Any(), min_rel=Any(), max_rel=Any(required=True)),
    ]

    def _evaluator(self, measures, qrels):
        invocations = {}
        measures = ir_measures.util.flatten_measures(measures)
        for measure in measures:
            if measure.NAME in ('P', 'RR', 'AP', 'RBP'):
                inv_key = (measure['rel'], None, None)
            elif measure.NAME in ('BPM', 'SDCG', 'NERR8', 'NERR9', 'NERR10', 'NERR11', 'INST', 'INSQ'):
                inv_key = (None, measure['min_rel'], measure['max_rel'])
            if inv_key not in invocations:
                invocations[inv_key] = []
            invocations[inv_key].append(measure)
        return CwlEvaluator(measures, qrels, invocations)

    def initialize(self):
        # disable the cwl logger (which writes to cwl.log)
        cwl_logger = logging.getLogger('cwl')
        cwl_logger.disabled = True


class IrmQrelHandler(TrecQrelHandler):
    def __init__(self, bin_rel_cutoff, min_rel, max_rel):
        super().__init__(None) # file name = None
        self.bin_rel_cutoff = bin_rel_cutoff
        self.min_rel = min_rel
        self.max_rel = max_rel
        if self.bin_rel_cutoff is None:
            assert self.min_rel is not None and self.max_rel is not None, "must provide either bin_rel_cutoff xor (BOTH min_rel and max_rel)"
            assert self.min_rel < self.max_rel, "min_rel must be less than max_rel"
        else:
            assert self.min_rel is None and self.max_rel is None, "must provide either bin_rel_cutoff xor (BOTH min_rel and max_rel)"
        self._min_observed_rel = float('inf')
        self._max_observed_rel = float('-inf')

    def read_file(self, qrels):
        pass # disable reading from file, we build these below

    def put_value(self, query_id, doc_id, relevance):
        if self.bin_rel_cutoff is not None:
            relevance = 1 if relevance >= self.bin_rel_cutoff else 0
        else:
            self._min_observed_rel = min(self._min_observed_rel, relevance)
            self._max_observed_rel = max(self._max_observed_rel, relevance)
            # clip value to range [min_rel, max_rel]
            relevance = min(max(relevance, self.min_rel), self.max_rel)
            # scale to be between [0, 1], based on the min_rel, max_rel range
            relevance = (relevance - self.min_rel) / (self.max_rel - self.min_rel)
        super().put_value(query_id, doc_id, relevance)

    def verify_gains(self):
        if self.bin_rel_cutoff is None:
            if self._min_observed_rel < self.min_rel:
                is_typical = self.min_rel == 0 and self._min_observed_rel < 0
                typical_message = ' This is typical and the desired behaviour in most TREC collections (where negative relevance scores are treated equally).' if is_typical else ''
                logger.warning(f'min_rel={self.min_rel} but at least one relevance score of {self._min_observed_rel} was observed. Scores less than {self.min_rel} were treated as {self.min_rel}.{typical_message}')
            if self._min_observed_rel > self.min_rel:
                logger.warning(f'min_rel={self.min_rel} but the lowest relevance score observed was {self._min_observed_rel}.')
            if self._max_observed_rel > self.max_rel:
                logger.warning(f'max_rel={self.max_rel} but at least one relevance score of {self._max_observed_rel} was observed. Scores greater than {self.max_rel} were treated as {self.max_rel}.')
            if self._max_observed_rel < self.max_rel:
                logger.warning(f'max_rel={self.max_rel} but at the highest relevance score observed was {self._max_observed_rel}. This is sometimes expected, e.g., if annotated on a scale up to {self._max_observed_rel} but no such documents were found.')


class CwlEvaluator(providers.Evaluator):
    def __init__(self, measures, qrels, invocations):
        self.qrhs = {}
        for inv_key in invocations.keys():
            self.qrhs[inv_key] = IrmQrelHandler(*inv_key)
        qids = set()
        for qrel in ir_measures.util.QrelsConverter(qrels).as_namedtuple_iter():
            qids.add(qrel.query_id)
            rel = max(qrel.relevance, 0) # clip all negative scores to 0, following trec_eval convention
            for qrh in self.qrhs.values():
                qrh.put_value(qrel.query_id, qrel.doc_id, rel)
        for qrh in self.qrhs.values():
            qrh.verify_gains()
        self.invocations = invocations
        super().__init__(measures, qids)

    def _iter_calc(self, run):
        # adapted from cwl_eval's main() method
        ranking_makers = None
        curr_qid = None
        for item in ir_measures.util.RunConverter(run).as_sorted_namedtuple_iter():
            if item.query_id not in self.qrel_qids:
                continue # skip queries not found in qrels; handled by base
            if item.query_id != curr_qid:
                if curr_qid is not None:
                    yield from self.flush(curr_qid, ranking_makers)
                curr_qid = item.query_id
                ranking_makers = {}
                for inv_key in self.invocations.keys():
                    ranking_makers[inv_key] = RankingMaker(curr_qid, self.qrhs[inv_key], min_gain=0., max_gain=1., cost_dict=None, max_cost=1.0, min_cost=1.0, max_n=1000) # max_n=1000 from cwl_eval default
            for maker in ranking_makers.values():
                maker.add(item.doc_id, "_")

        if curr_qid is not None:
            yield from self.flush(curr_qid, ranking_makers)

    def flush(self, qid, ranking_makers):
        rankings = {inv_key: maker.get_ranking() for inv_key, maker in ranking_makers.items()}
        for inv_key, measures in self.invocations.items():
            for measure in measures:
                cwl_measure = self._irm_convert_to_measure(measure)
                value = cwl_measure.measure(rankings[inv_key])
                yield CwlMetric(query_id=qid, measure=measure,
                    value=value,
                    expected_total_utility=cwl_measure.expected_total_utility,
                    expected_cost=cwl_measure.expected_cost,
                    expected_total_cost=cwl_measure.expected_total_cost,
                    expected_items=cwl_measure.expected_items)

    def _irm_convert_to_measure(self, measure):
        if measure.NAME == 'P':
            return PrecisionCWLMetric(measure['cutoff'])
        if measure.NAME == 'RR':
            return RRCWLMetric()
        if measure.NAME == 'AP':
            return APCWLMetric()
        if measure.NAME == 'RBP':
            return RBPCWLMetric(measure['p'])
        if measure.NAME == 'BPM':
            return BPMCWLMetric(measure['T'], measure['cutoff'])
        if measure.NAME == 'NERR8':
            return NERReq8CWLMetric(measure['cutoff'])
        if measure.NAME == 'NERR9':
            return NERReq9CWLMetric(measure['cutoff'])
        if measure.NAME == 'NERR10':
            return NERReq10CWLMetric(measure['p'])
        if measure.NAME == 'NERR11':
            return NERReq11CWLMetric(measure['T'])
        if measure.NAME == 'SDCG':
            return NDCGCWLMetric(measure['cutoff'])
        if measure.NAME == 'INST':
            return INSTCWLMetric(measure['T'])
        if measure.NAME == 'INSQ':
            return INSQCWLMetric(measure['T'])
        raise KeyError(f'measure {measure} not supported')


providers.register(CwlEvalProvider())



from ir_measures import measures
from .base import Measure, ParamInfo, SumAgg


class _INST(measures.Measure):
    """
    INST

::

     @inproceedings{Moffat:2012:MMI:2407085.2407092,
       author = {Moffat, Alistair and Scholer, Falk and Thomas, Paul},
       title = {Models and Metrics: IR Evaluation As a User Process},
       booktitle = {Proceedings of the Seventeenth Australasian Document Computing Symposium},
       year = {2012},
       url = {http://doi.acm.org/10.1145/2407085.2407092}
     }
    """
    __name__ = 'INST'
    NAME = __name__
    SUPPORTED_PARAMS = {
        'T': measures.ParamInfo(dtype=float, default=1.0, desc='TODO'),
        'min_rel': measures.ParamInfo(dtype=int, default=0, desc='minimum relevance score'),
        'max_rel': measures.ParamInfo(dtype=int, required=True, desc='maximum relevance score'),
    }


class _INSQ(measures.Measure):
    """
    INSQ, a variant of INST

::

     @inproceedings{Moffat:2015:IAM:2838931.2838938,
       author = {Moffat, Alistair and Bailey, Peter and Scholer, Falk and Thomas, Paul},
       title = {INST: An Adaptive Metric for Information Retrieval Evaluation},
       booktitle = {Proceedings of the 20th Australasian Document Computing Symposium},
       year = {2015},
       url = {http://doi.acm.org/10.1145/2838931.2838938}
     }
    """
    __name__ = 'INSQ'
    NAME = __name__
    SUPPORTED_PARAMS = {
        'T': measures.ParamInfo(dtype=float, default=1.0, desc='TODO'),
        'min_rel': measures.ParamInfo(dtype=int, default=0, desc='minimum relevance score'),
        'max_rel': measures.ParamInfo(dtype=int, required=True, desc='maximum relevance score'),
    }


INST = _INST()
measures.register(INST)

INSQ = _INSQ()
measures.register(INSQ)



from ir_measures import measures
from .base import Measure, ParamInfo


class _infAP(measures.Measure):
    """
    Inferred AP. AP implementation that accounts for pooled-but-unjudged documents by assuming
    that they are relevant at the same proportion as other judged documents. Essentially, skips
    documents that were pooled-but-not-judged, and assumes unjudged are non-relevant.

    Pooled-but-unjudged indicated by a score of -1, by convention. Note that not all qrels use
    this convention.
    """
    __name__ = 'infAP'
    NAME = __name__
    SUPPORTED_PARAMS = {
        'rel': measures.ParamInfo(dtype=int, default=1, desc='minimum relevance score to be considered relevant (inclusive)')
    }


infAP = _infAP()
measures.register(infAP)



from ir_measures import measures
from .base import Measure, ParamInfo


class _AP(measures.Measure):
    """
    The [Mean] Average Precision ([M]AP). The average precision of a single query is the mean
    of the precision scores at each relevant item returned in a search results list.
    
    AP is typically used for adhoc ranking tasks where getting as many relevant items as possible is. It is commonly referred to as MAP,
    by taking the mean of AP over the query set.

::

    @article{Harman:1992:ESIR,
      author = {Donna Harman},
      title = {Evaluation Issues in Information Retrieval},
      journal = {Information Processing and Management},
      volume = {28},
      number = {4},
      pages = {439 - -440},
      year = {1992},
    }
    """
    __name__ = 'AP'
    NAME = __name__
    SUPPORTED_PARAMS = {
        'cutoff': measures.ParamInfo(dtype=int, required=False, desc='ranking cutoff threshold'),
        'rel': measures.ParamInfo(dtype=int, default=1, desc='minimum relevance score to be considered relevant (inclusive)')
    }


AP = _AP()
MAP = AP
measures.register(AP, ['MAP'])



from ir_measures import measures
from .base import Measure, ParamInfo


class _NERR8(measures.Measure):
    """
    Version of the Not (but Nearly) Expected Reciprocal Rank (NERR) measure, version from Equation (8) of the the following paper.

::

     @inproceedings{Azzopardi:2021:ECE:3471158.3472239,
       author = {Azzopardi, Leif and Mackenzie, Joel and Moffat, Alistair},
       title = {{ERR} is not {C/W/L}: Exploring the Relationship Between Expected Reciprocal Rank and Other Metrics},
       booktitle = {ICTIR},
       year = {2021},
       url = {https://doi.org/10.1145/3471158.3472239}
     }
    """
    __name__ = 'NERR8'
    NAME = __name__
    SUPPORTED_PARAMS = {
        'cutoff': measures.ParamInfo(dtype=int, required=True, desc='ranking cutoff threshold'),
        'min_rel': measures.ParamInfo(dtype=int, default=0, desc='minimum relevance score'),
        'max_rel': measures.ParamInfo(dtype=int, required=True, desc='maximum relevance score'),
    }

class _NERR9(measures.Measure):
    """
    Version of the Not (but Nearly) Expected Reciprocal Rank (NERR) measure, version from Equation (9) of the the following paper.

::

     @inproceedings{Azzopardi:2021:ECE:3471158.3472239,
       author = {Azzopardi, Leif and Mackenzie, Joel and Moffat, Alistair},
       title = {{ERR} is not {C/W/L}: Exploring the Relationship Between Expected Reciprocal Rank and Other Metrics},
       booktitle = {ICTIR},
       year = {2021},
       url = {https://doi.org/10.1145/3471158.3472239}
     }
    """
    __name__ = 'NERR9'
    NAME = __name__
    SUPPORTED_PARAMS = {
        'cutoff': measures.ParamInfo(dtype=int, required=True, desc='ranking cutoff threshold'),
        'min_rel': measures.ParamInfo(dtype=int, default=0, desc='minimum relevance score'),
        'max_rel': measures.ParamInfo(dtype=int, required=True, desc='maximum relevance score'),
    }

class _NERR10(measures.Measure):
    """
    Version of the Not (but Nearly) Expected Reciprocal Rank (NERR) measure, version from Equation (10) of the the following paper.

::

     @inproceedings{Azzopardi:2021:ECE:3471158.3472239,
       author = {Azzopardi, Leif and Mackenzie, Joel and Moffat, Alistair},
       title = {{ERR} is not {C/W/L}: Exploring the Relationship Between Expected Reciprocal Rank and Other Metrics},
       booktitle = {ICTIR},
       year = {2021},
       url = {https://doi.org/10.1145/3471158.3472239}
     }
    """
    __name__ = 'NERR10'
    NAME = __name__
    SUPPORTED_PARAMS = {
        'p': measures.ParamInfo(dtype=float, default=0.9, desc='persistence'),
        'min_rel': measures.ParamInfo(dtype=int, default=0, desc='minimum relevance score'),
        'max_rel': measures.ParamInfo(dtype=int, required=True, desc='maximum relevance score'),
    }

class _NERR11(measures.Measure):
    """
    Version of the Not (but Nearly) Expected Reciprocal Rank (NERR) measure, version from Equation (12) of the the following paper.

::

     @inproceedings{Azzopardi:2021:ECE:3471158.3472239,
       author = {Azzopardi, Leif and Mackenzie, Joel and Moffat, Alistair},
       title = {{ERR} is not {C/W/L}: Exploring the Relationship Between Expected Reciprocal Rank and Other Metrics},
       booktitle = {ICTIR},
       year = {2021},
       url = {https://doi.org/10.1145/3471158.3472239}
     }
    """
    __name__ = 'NERR11'
    NAME = __name__
    SUPPORTED_PARAMS = {
        'T': measures.ParamInfo(dtype=float, default=1.0, desc='total desired gain (normalized)'),
        'min_rel': measures.ParamInfo(dtype=int, default=0, desc='minimum relevance score'),
        'max_rel': measures.ParamInfo(dtype=int, required=True, desc='maximum relevance score'),
    }


NERR8 = _NERR8()
measures.register(NERR8)
NERR9 = _NERR9()
measures.register(NERR9)
NERR10 = _NERR10()
measures.register(NERR10)
NERR11 = _NERR11()
measures.register(NERR11)



import ir_measures

def main():
    with open('docs/measures.rst', 'wt') as f:
        measures, aliases = [], []
        for name, val in ir_measures.measures.registry.items():
            if name == val.NAME:
                measures.append((name, val))
            else:
                aliases.append((name, val))
        measures = sorted(measures, key=lambda x: x[0].upper())
        aliases = sorted(aliases, key=lambda x: x[0].upper())
        f.write('''
Measures
=========================
''')
        for name, val in measures:
            f.write(f'''
``{name}``
-------------------------

{val.__doc__.replace('    ', '')}
''')
            if val.SUPPORTED_PARAMS:
                f.write('''**Parameters:**\n\n''')
                for p, param in val.SUPPORTED_PARAMS.items():
                    f.write(f'- ``{p}`` ({param.dtype.__name__}) - {param.desc}\n')
                f.write('\n\n')
            measure_providers = [(n, [m for m in p.SUPPORTED_MEASURES if m.__name__ == val.__name__]) for n, p in ir_measures.providers.registry.items()]
            measure_providers = [(n, m) for n, m in measure_providers if m]
            f.write('''**Provided by:**\n\n''')
            if measure_providers:
                for p, ms in measure_providers:
                    ms = ', '.join(f'``{m}``' for m in ms)
                    f.write(f'- ``{p}``: {ms}\n')
                f.write('\n\n')
            else:
                f.write('\n\n*(none yet)*\n\n')
        f.write('''
Aliases
-------------------------

These provide shortcuts to "canonical" measures, and are typically used when multiple
names or casings for the same measure exist. You can use them just like any other measure
and the identifiers are equal (e.g., ``AP == MAP``) but the names will appear in the
canonical form when printed.

''')
        for name, val in aliases:
            f.write(f'- ``{name}`` → ``{val}``\n')

    with open('docs/providers.rst', 'wt') as f:
        f.write('''
Providers
=========================
''')
        for name, val in sorted(ir_measures.providers.registry.items()):
            f.write(f'''
``{name}``
-------------------------

{val.__doc__.replace('    ', ' ')}
''')
            f.write('''**Supported Measures:**\n\n''')
            for measure in val.SUPPORTED_MEASURES:
                f.write(f' - ``{measure}``\n')
            f.write('\n\n')

if __name__ == '__main__':
    main()



from ir_measures import measures
from .base import Measure, ParamInfo


class _Success(measures.Measure):
    """
    1 if a document with at least rel relevance is found in the first cutoff documents, else 0.

    NOTE: Some refer to this measure as Recall@k. This software follows the TREC convention, where
    Recall@k is defined as the proportion of known relevant documents retrieved in the top k results.
    """
    __name__ = 'Success'
    NAME = __name__
    SUPPORTED_PARAMS = {
        'cutoff': measures.ParamInfo(dtype=int, required=True, desc='ranking cutoff threshold'),
        'rel': measures.ParamInfo(dtype=int, default=1, desc='minimum relevance score to be considered relevant (inclusive)')
    }


Success = _Success()
measures.register(Success)



import contextlib
import ir_measures
from ir_measures import providers, measures, Metric
from ir_measures.providers.base import Any, Choices, NOT_PROVIDED


class PytrecEvalProvider(providers.Provider):
    """
    pytrec_eval

    https://github.com/cvangysel/pytrec_eval

::

    @inproceedings{VanGysel2018pytreceval,
        title={Pytrec\\_eval: An Extremely Fast Python Interface to trec\\_eval},
        author={Van Gysel, Christophe and de Rijke, Maarten},
        publisher={ACM},
        booktitle={SIGIR},
        year={2018},
    }

    """
    NAME = 'pytrec_eval'
    SUPPORTED_MEASURES = [
        measures._P(cutoff=Any(), rel=Any()),
        measures._RR(cutoff=Choices(NOT_PROVIDED), rel=Any()),
        measures._Rprec(rel=Any()),
        measures._AP(cutoff=Any(), rel=Any()),
        measures._nDCG(cutoff=Any(), dcg=Choices('log2')),
        measures._R(cutoff=Any()),
        measures._Bpref(rel=Any()),
        measures._NumRet(rel=Any()),
        measures._NumQ(),
        measures._NumRel(rel=Choices(1)), # for some reason, relevance_level doesn't flow through to num_rel, so can only support rel=1
        measures._SetAP(rel=Any()),
        measures._SetF(rel=Any(), beta=Any()),
        measures._SetP(rel=Any(), relative=Any()),
        measures._SetR(rel=Any()),
        measures._Success(rel=Any(), cutoff=Any()),
        measures._IPrec(recall=Any()),
        measures._infAP(rel=Any()),
        # Cannot support Judged because software doesn't support negative relevance levels: <https://github.com/cvangysel/pytrec_eval/blob/2362660e02c324df281932cc23ad7efd31cd3957/src/pytrec_eval.cpp#L354>
    ]

    def __init__(self):
        super().__init__()
        self.pytrec_eval = None

    def _evaluator(self, measures, qrels):
        measures = ir_measures.util.flatten_measures(measures)
        # Convert qrels to dict_of_dict (input format used by pytrec_eval)
        qrels = ir_measures.util.QrelsConverter(qrels).as_dict_of_dict()

        # Depending on the measure params, we may need multiple invocations of pytrec_eval
        # (e.g., with different rel_level, since it only supports running with 1 rel_level at a time)
        invokers = self._build_invokers(measures, qrels)
        return PytrecEvalEvaluator(measures, invokers, qrels)

    def _build_invokers(self, measures, qrels):
        invocations = {}
        setf_count = 0
        for measure in measures:
            match_str = None
            if measure.NAME == 'P':
                invocation_key = (measure['rel'], 0)
                measure_str = f'P_{measure["cutoff"]}'
            elif measure.NAME == 'RR':
                invocation_key = (measure['rel'], 0)
                measure_str = f'recip_rank'
            elif measure.NAME == 'Rprec':
                invocation_key = (measure['rel'], 0)
                measure_str = f'Rprec'
            elif measure.NAME == 'AP':
                invocation_key = (measure['rel'], 0)
                if measure['cutoff'] is NOT_PROVIDED:
                    measure_str = f'map'
                else:
                    measure_str = f'map_cut_{measure["cutoff"]}'
            elif measure.NAME == 'infAP':
                invocation_key = (measure['rel'], 0)
                measure_str = f'infAP'
            elif measure.NAME == 'nDCG':
                # Doesn't matter where this goes... Put it in an existing invocation, or just (1,) if none yet exist
                if invocations:
                    invocation_key = next(iter(invocations))
                else:
                    invocation_key = (1, 0)
                if measure['cutoff'] is NOT_PROVIDED:
                    measure_str = f'ndcg'
                else:
                    measure_str = f'ndcg_cut_{measure["cutoff"]}'
            elif measure.NAME == 'R':
                invocation_key = (measure['rel'], 0)
                measure_str = f'recall_{measure["cutoff"]}'
            elif measure.NAME == 'Bpref':
                invocation_key = (measure['rel'], 0)
                measure_str = f'bpref'
            elif measure.NAME == 'NumRet':
                if measure['rel'] is NOT_PROVIDED:
                    # Doesn't matter where this goes... Put it in an existing invocation, or just (1,) if none yet exist
                    if invocations:
                        invocation_key = next(iter(invocations))
                    else:
                        invocation_key = (1, 0)
                    measure_str = 'num_ret'
                else:
                    invocation_key = (measure['rel'], 0)
                    measure_str = 'num_rel_ret'
            elif measure.NAME == 'NumQ':
                # Doesn't matter where this goes... Put it in an existing invocation, or just (1,) if none yet exist
                if invocations:
                    invocation_key = next(iter(invocations))
                else:
                    invocation_key = (1, 0)
                measure_str = 'num_q'
            elif measure.NAME == 'NumRel':
                invocation_key = (measure['rel'], 0)
                measure_str = 'num_rel'
            elif measure.NAME == 'SetAP':
                invocation_key = (measure['rel'], 0)
                measure_str = f'set_map'
            elif measure.NAME == 'SetF':
                # set_F is strange (or buggy?) in both trec_eval and pytrec_eval. It only accepts
                # the first beta argument it's given, which is why we use the setf_count approach
                # to handle multiple invocations. It also is always reported as the name set_F by
                # pytrec_eval, so we need different measure_str and match_str here.
                invocation_key = (measure['rel'], setf_count)
                setf_count += 1
                measure_str = f'set_F_{measure["beta"]}'
                match_str = 'set_F'
                if measure['beta'] == 1.:
                    measure_str = f'set_F'
                else:
                    measure_str = f'set_F_{measure["beta"]}'
            elif measure.NAME == 'SetP':
                if measure['relative']:
                    invocation_key = (measure['rel'], 0)
                    measure_str = f'set_relative_P'
                else:
                    invocation_key = (measure['rel'], 0)
                    measure_str = f'set_P'
            elif measure.NAME == 'SetR':
                invocation_key = (measure['rel'], 0)
                measure_str = f'set_recall'
            elif measure.NAME == 'Success':
                invocation_key = (measure['rel'], 0)
                measure_str = f'success_{measure["cutoff"]}'
            elif measure.NAME == 'IPrec':
                invocation_key = (measure['rel'], 0)
                measure_str = f'iprec_at_recall_{measure["recall"]:.2f}'
            else:
                raise ValueError(f'unsupported measure {measure}')

            if match_str is None:
                match_str = measure_str

            if invocation_key not in invocations:
                invocations[invocation_key] = {}
            invocations[invocation_key][match_str] = (measure, measure_str)

        invokers = []
        for (rel_level, it), measure_map in invocations.items():
            invokers.append(PytrecEvalInvoker(self.pytrec_eval, qrels, measure_map, rel_level))

        return invokers

    def initialize(self):
        try:
            import pytrec_eval
            self.pytrec_eval = pytrec_eval
        except ImportError as ex:
            raise RuntimeError('pytrec_eval not available', ex)


class PytrecEvalEvaluator(providers.Evaluator):
    def __init__(self, measures, invokers, qrels):
        super().__init__(measures, set(qrels.keys()))
        self.invokers = invokers

    def _iter_calc(self, run):
        # Convert qrels to dict_of_dict (input format used by pytrec_eval)
        run = ir_measures.util.RunConverter(run).as_dict_of_dict()
        for invoker in self.invokers:
            yield from invoker.iter_calc(run)


class PytrecEvalInvoker:
    def __init__(self, pte, qrels, measure_map, rel_level):
        self.evaluator = pte.RelevanceEvaluator(qrels, [m for _, m in measure_map.values()], relevance_level=rel_level)
        self.measure_map = measure_map

    def iter_calc(self, run):
        result = self.evaluator.evaluate(run)
        for query_id, measures in result.items():
            for measure_str, value in measures.items():
                yield Metric(query_id=query_id, measure=self.measure_map[measure_str][0], value=value)


providers.register(PytrecEvalProvider())



from ir_measures import measures
from .base import BaseMeasure, ParamInfo


class _ERR_IA(measures.BaseMeasure):
    """
    Intent-Aware Expected Reciprocal Rank with collection-independent normalisation.

::

    @inproceedings{10.1145/1645953.1646033,
      author = {Chapelle, Olivier and Metlzer, Donald and Zhang, Ya and Grinspan, Pierre},
      title = {Expected Reciprocal Rank for Graded Relevance},
      booktitle = {CIKM},
      year = {2009}
    }
    """
    __name__ = 'ERR_IA'
    NAME = __name__
    SUPPORTED_PARAMS = {
        'cutoff': measures.ParamInfo(dtype=int, required=False, desc='ranking cutoff threshold'),
        'rel': measures.ParamInfo(dtype=int, default=1, desc='minimum relevance score to be considered relevant (inclusive)'),
        'judged_only': measures.ParamInfo(dtype=bool, default=False, desc='calculate measure using only judged documents (i.e., discard unjudged documents)'),
    }

class _nERR_IA(measures.BaseMeasure):
    """
    Intent-Aware Expected Reciprocal Rank with collection-dependent normalisation.

::

    @inproceedings{10.1145/1645953.1646033,
      author = {Chapelle, Olivier and Metlzer, Donald and Zhang, Ya and Grinspan, Pierre},
      title = {Expected Reciprocal Rank for Graded Relevance},
      booktitle = {CIKM},
      year = {2009}
    }
    """
    __name__ = 'nERR_IA'
    NAME = __name__
    SUPPORTED_PARAMS = {
        'cutoff': measures.ParamInfo(dtype=int, required=False, desc='ranking cutoff threshold'),
        'rel': measures.ParamInfo(dtype=int, default=1, desc='minimum relevance score to be considered relevant (inclusive)'),
        'judged_only': measures.ParamInfo(dtype=bool, default=False, desc='calculate measure using only judged documents (i.e., discard unjudged documents)'),
    }

class _alpha_DCG(measures.BaseMeasure):
    """
    A version of DCG that accounts for multiple possible query intents.

::

    @inproceedings{Clarke2008NoveltyAD,
      title={Novelty and diversity in information retrieval evaluation},
      author={Charles L. A. Clarke and Maheedhar Kolla and Gordon V. Cormack and Olga Vechtomova and Azin Ashkan and Stefan B{\"u}ttcher and Ian MacKinnon},
      booktitle={SIGIR},
      year={2008}
    }
    """
    __name__ = 'alpha_DCG'
    NAME = __name__
    SUPPORTED_PARAMS = {
        'cutoff': measures.ParamInfo(dtype=int, required=False, desc='ranking cutoff threshold'),
        'rel': measures.ParamInfo(dtype=int, default=1, desc='minimum relevance score to be considered relevant (inclusive)'),
        'alpha': measures.ParamInfo(dtype=float, default=0.5, desc='Redundancy intolerance'),
        'judged_only': measures.ParamInfo(dtype=bool, default=False, desc='calculate measure using only judged documents (i.e., discard unjudged documents)'),
    }

class _alpha_nDCG(measures.BaseMeasure):
    """
    A version of nDCG that accounts for multiple possible query intents.

::

    @inproceedings{Clarke2008NoveltyAD,
      title={Novelty and diversity in information retrieval evaluation},
      author={Charles L. A. Clarke and Maheedhar Kolla and Gordon V. Cormack and Olga Vechtomova and Azin Ashkan and Stefan B{\"u}ttcher and Ian MacKinnon},
      booktitle={SIGIR},
      year={2008}
    }
    """
    __name__ = 'alpha_nDCG'
    NAME = __name__
    SUPPORTED_PARAMS = {
        'cutoff': measures.ParamInfo(dtype=int, required=False, desc='ranking cutoff threshold'),
        'rel': measures.ParamInfo(dtype=int, default=1, desc='minimum relevance score to be considered relevant (inclusive)'),
        'alpha': measures.ParamInfo(dtype=float, default=0.5, desc='Redundancy intolerance'),
        'judged_only': measures.ParamInfo(dtype=bool, default=False, desc='calculate measure using only judged documents (i.e., discard unjudged documents)'),
    }

class _NRBP(measures.BaseMeasure):
    """
    Novelty- and Rank-Biased Precision with collection-independent normalisation.

::

    @InProceedings{10.1007/978-3-642-04417-5_17,
      author="Clarke, Charles L. A. and Kolla, Maheedhar and Vechtomova, Olga",
      title="An Effectiveness Measure for Ambiguous and Underspecified Queries ",
      booktitle="ICTIR",
      year="2009"
    }
    """
    __name__ = 'NRBP'
    NAME = __name__
    SUPPORTED_PARAMS = {
        'rel': measures.ParamInfo(dtype=int, default=1, desc='minimum relevance score to be considered relevant (inclusive)'),
        'alpha': measures.ParamInfo(dtype=float, default=0.5, desc='Redundancy intolerance'),
        'beta': measures.ParamInfo(dtype=float, default=0.5, desc='Patience'),
    }

class _nNRBP(measures.BaseMeasure):
    """
    Novelty- and Rank-Biased Precision with collection-dependent normalisation.

::

    @InProceedings{10.1007/978-3-642-04417-5_17,
      author="Clarke, Charles L. A. and Kolla, Maheedhar and Vechtomova, Olga",
      title="An Effectiveness Measure for Ambiguous and Underspecified Queries ",
      booktitle="ICTIR",
      year="2009"
    }
    """
    __name__ = 'nNRBP'
    NAME = __name__
    SUPPORTED_PARAMS = {
        'rel': measures.ParamInfo(dtype=int, default=1, desc='minimum relevance score to be considered relevant (inclusive)'),
        'alpha': measures.ParamInfo(dtype=float, default=0.5, desc='Redundancy intolerance'),
        'beta': measures.ParamInfo(dtype=float, default=0.5, desc='Patience'),
    }

class _AP_IA(measures.BaseMeasure):
    """
    Intent-aware (Mean) Average Precision
    """
    __name__ = 'AP_IA'
    NAME = __name__
    SUPPORTED_PARAMS = {
        'rel': measures.ParamInfo(dtype=int, default=1, desc='minimum relevance score to be considered relevant (inclusive)'),
        'judged_only': measures.ParamInfo(dtype=bool, default=False, desc='calculate measure using only judged documents (i.e., discard unjudged documents)'),
    }

class _P_IA(measures.BaseMeasure):
    """
    Intent-aware Precision@k.
    """
    __name__ = 'P_IA'
    NAME = __name__
    SUPPORTED_PARAMS = {
        'cutoff': measures.ParamInfo(dtype=int, required=False, desc='ranking cutoff threshold'),
        'rel': measures.ParamInfo(dtype=int, default=1, desc='minimum relevance score to be considered relevant (inclusive)'),
        'judged_only': measures.ParamInfo(dtype=bool, default=False, desc='calculate measure using only judged documents (i.e., discard unjudged documents)'),
    }

class _StRecall(measures.BaseMeasure):
    """
    Subtopic recall (the number of subtopics covered by the top k docs)
    """
    __name__ = 'StRecall'
    NAME = __name__
    SUPPORTED_PARAMS = {
        'cutoff': measures.ParamInfo(dtype=int, required=False, desc='ranking cutoff threshold'),
        'rel': measures.ParamInfo(dtype=int, default=1, desc='minimum relevance score to be considered relevant (inclusive)'),
    }



ERR_IA = _ERR_IA()
nERR_IA = _nERR_IA()
alpha_DCG = _alpha_DCG()
α_DCG = alpha_DCG
alpha_nDCG = _alpha_nDCG()
α_nDCG = alpha_nDCG
NRBP = _NRBP()
nNRBP = _nNRBP()
AP_IA = _AP_IA()
MAP_IA = AP_IA
P_IA = _P_IA()
StRecall = _StRecall()

measures.register(ERR_IA)
measures.register(nERR_IA)
measures.register(alpha_DCG, aliases=['α_DCG'])
measures.register(alpha_nDCG, aliases=['α_nDCG'])
measures.register(NRBP)
measures.register(nNRBP)
measures.register(AP_IA, aliases=['MAP_IA'])
measures.register(P_IA)
measures.register(StRecall)



__version__ = "0.2.3"
import sys
import logging
from . import util
from . import lazylibs
from .util import (parse_measure, parse_trec_measure,
                   read_trec_qrels, read_trec_run,
                   Qrel, ScoredDoc, Metric,
                   GenericQrel, # deprecated; replaced with Qrel
                   GenericScoredDoc, # deprecated; replaced with ScoredDoc
                   convert_trec_name, # deprecated; replaced with parse_trec_measure
                   parse_trec_qrels, # deprecated; replaced with read_trec_qrels
                   parse_trec_run, # deprecated; replaced with read_trec_run
                  )
from . import measures
from .measures import *
from .measures import Measure
from . import providers

logger = logging.getLogger('ir_measures')
logger.setLevel('WARNING')
log_handler = logging.StreamHandler(sys.stderr)
log_handler.setFormatter(logging.Formatter('[%(name)s] [%(levelname)s] %(message)s'))
logger.addHandler(log_handler)

# providers
cwl_eval = providers.registry['cwl_eval']
compat = providers.registry['compat']
gdeval = providers.registry['gdeval']
pytrec_eval = providers.registry['pytrec_eval']
trectools = providers.registry['trectools']
judged = providers.registry['judged']
msmarco = providers.registry['msmarco']
pyndeval = providers.registry['pyndeval']

CwlMetric = providers.CwlMetric

DefaultPipeline = providers.FallbackProvider([
    pytrec_eval,
    cwl_eval,
    compat,
    pyndeval,
    # trectools,  # buggy; will add back later
    judged,
    msmarco,
    gdeval,  # doesn't work when installed from package #9
])
evaluator = DefaultPipeline.evaluator
calc_ctxt = DefaultPipeline.calc_ctxt # deprecated; replaced with evaluator
iter_calc = DefaultPipeline.iter_calc
calc_aggregate = DefaultPipeline.calc_aggregate

__all__ = measures.__all__



import pandas as pd
import unittest
import itertools
import ir_measures
from ir_measures import *
from pandas.testing import assert_frame_equal

class TestUtil(unittest.TestCase):

    def test_parse_trec_measure(self):
        cases = {
            'map': [AP],
            'P_5': [P@5],
            'P_5,10': [P@5, P@10],
            'P.5': [P@5],
            'P.5,10': [P@5, P@10],
            'P': [P@5, P@10, P@15, P@20, P@30, P@100, P@200, P@500, P@1000],
            'ndcg': [nDCG],
            'ndcg_cut.10': [nDCG@10],
            'ndcg_cut.5,10': [nDCG@5, nDCG@10],
            'ndcg_cut_10': [nDCG@10],
            'ndcg_cut_5,10': [nDCG@5, nDCG@10],
            'ndcg_cut': [nDCG@5, nDCG@10, nDCG@15, nDCG@20, nDCG@30, nDCG@100, nDCG@200, nDCG@500, nDCG@1000],
            'set_P': [SetP],
            'set_recall': [SetR],
            'set_F': [SetF],
            'set_relative_P': [SetRelP],
            'set_map': [SetAP],
            'set_F.1.0,0.5,2.4': [SetF, SetF(beta=0.5), SetF(beta=2.4)],
            'official': [P@5, P@10, P@15, P@20, P@30, P@100, P@200, P@500, P@1000, Rprec, Bpref, IPrec@0.0, IPrec@0.1, IPrec@0.2, IPrec@0.3, IPrec@0.4, IPrec@0.5, IPrec@0.6, IPrec@0.7, IPrec@0.8, IPrec@0.9, IPrec@1.0, AP, NumQ, NumRel, NumRelRet, NumRet, RR],
        }
        for case in cases:
            with self.subTest(case):
                self.assertEqual(ir_measures.parse_trec_measure(case), cases[case])

    def test_parse_measure(self):
        tests = {
            'AP': AP,
            AP: AP,
            'MAP': AP,
            MAP: MAP,
            'P@10': P@10,
            P@10: P@10,
            'nDCG@10': nDCG@10,
            'P(rel=2)@10': P(rel=2)@10,
            'nDCG(dcg="exp-log2")@10': nDCG(dcg='exp-log2')@10,
            'nDCG(dcg="exp-log2", cutoff=20)@10': nDCG(dcg='exp-log2')@10,
            'nDCG(dcg="exp-log2", cutoff=20)': nDCG(dcg='exp-log2')@20,
            'IPrec@0.2': IPrec@0.2,
            'IPrec(rel=2)@0.2': IPrec(rel=2)@0.2,
            'IPrec(rel=2, recall=0.4)@0.2': IPrec(rel=2)@0.2,
            'IPrec(rel=2, recall=0.4)': IPrec(rel=2)@0.4,
            IPrec(rel=2)@0.4: IPrec(rel=2)@0.4,
        }
        for key, value in tests.items():
            with self.subTest(key):
                self.assertEqual(ir_measures.parse_measure(key), value)

    def test_qrels_converter(self):
        qrels_list = [
            ir_measures.Qrel('1', 'A', 1),
            ir_measures.Qrel('1', 'B', 0),
            ir_measures.Qrel('2', 'A', 0),
            ir_measures.Qrel('2', 'C', 1),
        ]
        qrels_df = pd.DataFrame(qrels_list)
        qrels_df_noit = qrels_df.drop(columns='iteration')
        qrels_dict = {
            '1': {'A': 1, 'B': 0},
            '2': {'A': 0, 'C': 1}
        }
        sources = {
            'qrels_nt_list': lambda: qrels_list,
            'qrels_nt_iter': lambda: iter(qrels_list),
            'qrels_df': lambda: qrels_df,
            'qrels_df_noit': lambda: qrels_df_noit,
            'qrels_dict': lambda: qrels_dict
        }
        for n, fn in sources.items():
            with self.subTest(n):
                self.assertEqual(ir_measures.util.QrelsConverter(fn()).as_dict_of_dict(), qrels_dict)
                self.assertEqual(list(ir_measures.util.QrelsConverter(fn()).as_namedtuple_iter()), qrels_list)
                assert_frame_equal(ir_measures.util.QrelsConverter(fn()).as_pd_dataframe(), qrels_df)

        bad_df = pd.DataFrame([
            {'query_id': '0', 'docno': 'A', 'relevance': 1}
        ])
        with self.assertRaises(ValueError) as context:
            ir_measures.util.QrelsConverter(bad_df).as_dict_of_dict()

        self.assertEqual(context.exception.args[0], "unknown qrels format: DataFrame missing columns: ['doc_id'] (found ['query_id', 'docno', 'relevance'])")

        with self.assertRaises(ValueError) as context:
            ir_measures.util.QrelsConverter(object()).as_dict_of_dict()

        self.assertEqual(context.exception.args[0], "unknown qrels format: unexpected format; please provide either: (1) an iterable of namedtuples (fields ('query_id', 'doc_id', 'relevance'), e.g., from ir_measures.Qrel); (2) a pandas DataFrame with columns ('query_id', 'doc_id', 'relevance'); or (3) a dict-of-dict")

    def test_run_converter(self):
        run_list = [
            ir_measures.ScoredDoc('1', 'A', 1.2),
            ir_measures.ScoredDoc('1', 'B', 0.9),
            ir_measures.ScoredDoc('2', 'A', 3.5),
            ir_measures.ScoredDoc('2', 'C', 0.3),
        ]
        run_df = pd.DataFrame(run_list)
        run_dict = {
            '1': {'A': 1.2, 'B': 0.9},
            '2': {'A': 3.5, 'C': 0.3}
        }
        sources = {
            'run_nt_list': lambda: run_list,
            'run_nt_iter': lambda: iter(run_list),
            'run_df': lambda: run_df,
            'run_dict': lambda: run_dict
        }
        for n, fn in sources.items():
            with self.subTest(n):
                self.assertEqual(ir_measures.util.RunConverter(fn()).as_dict_of_dict(), run_dict)
                self.assertEqual(list(ir_measures.util.RunConverter(fn()).as_namedtuple_iter()), run_list)
                assert_frame_equal(ir_measures.util.RunConverter(fn()).as_pd_dataframe(), run_df)

        bad_df = pd.DataFrame([
            {'query_id': '0', 'docno': 'A', 'score': 1.2}
        ])
        with self.assertRaises(ValueError) as context:
            ir_measures.util.RunConverter(bad_df).as_dict_of_dict()

        self.assertEqual(context.exception.args[0], "unknown run format: DataFrame missing columns: ['doc_id'] (found ['query_id', 'docno', 'score'])")

        with self.assertRaises(ValueError) as context:
            ir_measures.util.RunConverter(object()).as_dict_of_dict()

        self.assertEqual(context.exception.args[0], "unknown run format: unexpected format; please provide either: (1) an iterable of namedtuples (fields ('query_id', 'doc_id', 'score'), e.g., from ir_measures.ScoredDoc); (2) a pandas DataFrame with columns ('query_id', 'doc_id', 'score'); or (3) a dict-of-dict")


if __name__ == '__main__':
    unittest.main()



from ir_measures import measures
from .base import Measure, ParamInfo


class _Compat(measures.Measure):
    """
    Compatibility measure desribed in:

::

    @article{10.1145/3451161,
      author = {Clarke, Charles L. A. and Vtyurina, Alexandra and Smucker, Mark D.},
      title = {Assessing Top-k Preferences},
      journal = {ACM Transactions on Information Systems},
      volume = {39},
      number = {3},
      articleno = {33},
      numpages = {21},
      year = {2021},
      url = {https://doi.org/10.1145/3451161},
    }
    """
    __name__ = 'Compat'
    NAME = __name__
    SUPPORTED_PARAMS = {
        'p': measures.ParamInfo(dtype=float, default=0.95, desc='persistence'),
        'normalize': measures.ParamInfo(dtype=bool, default=True, desc='apply normalization for finite ideal rankings'),
    }


Compat = _Compat()
measures.register(Compat)



from ir_measures import measures
from .base import Measure, ParamInfo


class _RR(measures.Measure):
    """
    The [Mean] Reciprocal Rank ([M]RR) is a precision-focused measure that scores based on the reciprocal of the rank of the
    highest-scoring relevance document. An optional cutoff can be provided to limit the
    depth explored. rel (default 1) controls which relevance level is considered relevant.

::

    @article{kantor2000trec,
      title={The TREC-5 Confusion Track},
      author={Kantor, Paul and Voorhees, Ellen},
      journal={Information Retrieval},
      volume={2},
      number={2-3},
      pages={165--176},
      year={2000}
    }
    """
    __name__ = 'RR'
    NAME = __name__
    SUPPORTED_PARAMS = {
        'cutoff': measures.ParamInfo(dtype=int, required=False, desc='ranking cutoff threshold'),
        'rel': measures.ParamInfo(dtype=int, default=1, desc='minimum relevance score to be considered relevant (inclusive)')
    }


RR = _RR()
MRR = RR
measures.register(RR, ['MRR'])



from glob import glob
import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

def read(rel_path):
    import os
    import codecs
    here = os.path.abspath(os.path.dirname(__file__))
    with codecs.open(os.path.join(here, rel_path), 'r') as fp:
        return fp.read()

def get_version(rel_path):
    import os
    suffix = os.environ["VERSION_SUFFIX" ] if "VERSION_SUFFIX" in os.environ else ""
    for line in read(rel_path).splitlines():
        if line.startswith('__version__'):
            delim = '"' if '"' in line else "'"
            return line.split(delim)[1] + suffix
    else:
        raise RuntimeError("Unable to find version string.")

setuptools.setup(
    name="ir_measures",
    version=get_version("ir_measures/__init__.py"),
    author="Sean MacAvaney",
    author_email="sean.macavaney@glasgow.ac.uk",
    description="provides a common interface to many IR measure tools",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/terrierteam/ir_measures",
    include_package_data = True,
    packages=setuptools.find_packages(include=['ir_measures', 'ir_measures.*']),
    install_requires=list(open('requirements.txt')),
    extras_require={
        "trectools": ["trectools>=0.0.44"],
    },
    classifiers=[],
    python_requires='>=3.6',
    entry_points={
        'console_scripts': ['ir_measures=ir_measures.__main__:main_cli'],
    },
    package_data={
        'ir_measures': ['ir_measures/bin/gdeval.pl', 'LICENSE.txt', 'requirements.txt', 'requirements-test.txt'],
    },
)



from ir_measures import measures
from .base import Measure, ParamInfo


class _BPM(measures.Measure):
    """
    The Bejeweled Player Model (BPM).

::

     @inproceedings{Zhang:2017:EWS:3077136.3080841,
       author = {Zhang, Fan and Liu, Yiqun and Li, Xin and Zhang, Min and Xu, Yinghui and Ma, Shaoping},
       title = {Evaluating Web Search with a Bejeweled Player Model},
       booktitle = {SIGIR},
       year = {2017},
       url = {http://doi.acm.org/10.1145/3077136.3080841}
     }
    """
    __name__ = 'BPM'
    NAME = __name__
    SUPPORTED_PARAMS = {
        'cutoff': measures.ParamInfo(dtype=int, required=True, desc='ranking cutoff threshold'),
        'T': measures.ParamInfo(dtype=float, default=1., desc='total desired gain (normalized)'),
        'min_rel': measures.ParamInfo(dtype=int, default=0, desc='minimum relevance score'),
        'max_rel': measures.ParamInfo(dtype=int, required=True, desc='maximum relevance score'),
    }


BPM = _BPM()
measures.register(BPM)



import warnings
import contextlib
import itertools
from typing import Iterator, Dict, Union
from ir_measures import providers, measures, Metric


class Evaluator:
    """
    The base class for scoring runs for a given set of measures and qrels.
    Returned from ``.evaluator(measures, qrels)`` calls.
    """
    def __init__(self, measures, qrel_qids):
        self.measures = measures
        self.qrel_qids = qrel_qids

    def iter_calc(self, run) -> Iterator['Metric']:
        """
        Yields per-topic metrics this run.
        """
        expected_measure_qids = set(itertools.product(self.measures, self.qrel_qids))
        for metric in self._iter_calc(run):
            expected_measure_qids.discard((metric.measure, metric.query_id))
            yield metric
        for measure, query_id in sorted(expected_measure_qids, key=lambda x: (str(x[0]), x[1])):
            yield Metric(query_id=query_id, measure=measure, value=measure.DEFAULT)

    def _iter_calc(self, run) -> Iterator['Metric']:
        raise NotImplementedError()

    def calc_aggregate(self, run) -> Dict[measures.Measure, Union[float, int]]:
        """
        Returns aggregated measure values for this run.
        """
        aggregators = {m: m.aggregator() for m in self.measures}
        for metric in self.iter_calc(run):
            aggregators[metric.measure].add(metric.value)
        return {m: agg.result() for m, agg in aggregators.items()}


class Provider:
    """
    The base class for all measure providers (e.g., pytrec_eval, gdeval, etc.).
    """
    NAME = None
    SUPPORTED_MEASURES = []

    def __init__(self):
        self._is_available = None

    def evaluator(self, measures, qrels) -> Evaluator:
        if self.is_available():
            return self._evaluator(measures, qrels)
        else:
            raise RuntimeError('provider not available')

    @contextlib.contextmanager
    def calc_ctxt(self, measures, qrels):
        warnings.warn("calc_ctxt deprecated in 0.2.0. Please use .evaluator(measures, qrels) instead.", DeprecationWarning)
        if self.is_available():
            evaluator = self._evaluator(measures, qrels)
            def _eval(run):
                yield from evaluator.iter_calc(run)
            yield _eval
        else:
            raise RuntimeError('provider not available')

    def iter_calc(self, measures, qrels, run):
        if self.is_available():
            return self._iter_calc(measures, qrels, run)
        else:
            raise RuntimeError('provider %s not available' % self.NAME)

    def _evaluator(self, measures, qrels):
        raise NotImplementedError()

    def _iter_calc(self, measures, qrels, run):
        return self._evaluator(measures, qrels).iter_calc(run)

    def calc_aggregate(self, measures, qrels, run):
        return self.evaluator(measures, qrels).calc_aggregate(run)

    def supports(self, measure):
        measure.validate_params()
        for supported_measure in self.SUPPORTED_MEASURES:
            if supported_measure.NAME == measure.NAME:
                valid = True
                for param_name, param_spec in supported_measure.params.items():
                    if not param_spec.validate(measure[param_name]):
                        valid = False
                        break
                if valid:
                    return True
        return False

    def is_available(self):
        if self._is_available is not None:
            return self._is_available
        try:
            self.initialize()
            self._is_available = True
        except RuntimeError:
            self._is_available = False
        return self._is_available

    def initialize(self):
        pass


class ParamSpec:
    def validate(self, value):
        raise NotImplementedError()

class Any:
    def __init__(self, required=False):
        self.required = required

    def validate(self, value):
        if self.required:
            return value is not NOT_PROVIDED
        return True

    def __repr__(self):
        if not self.required:
            return 'ANY'
        return 'REQUIRED'

class Choices:
    def __init__(self, *args):
        self.choices = args

    def validate(self, value):
        return value in self.choices

    def __repr__(self):
        if len(self.choices) == 1:
            if self.choices[0] == NOT_PROVIDED:
                return 'NOT_PROVIDED'
            return repr(self.choices[0])
        return repr(self.choices)

NOT_PROVIDED = measures.base._NOT_PROVIDED
